@@ -1,14 +1,16 @@
 # SPDX-License-Identifier: BSD-3-Clause
-from .client import Client, ProtocolVersion, TLSParameters, Will
+from ._version import __version__, __version_tuple__
+from .client import Client, ProtocolVersion, ProxySettings, TLSParameters, Will
 from .error import MqttCodeError, MqttError
-from .version import __version__
 
 __all__ = [
-    "MqttError",
-    "MqttCodeError",
-    "Client",
-    "Will",
-    "ProtocolVersion",
-    "TLSParameters",
     "__version__",
+    "__version_tuple__",
+    "Client",
+    "ProtocolVersion",
+    "ProxySettings",
+    "TLSParameters",
+    "Will",
+    "MqttCodeError",
+    "MqttError",
 ]
